//
use actix::prelude::*;

use git_next_core::{git, RepoConfigSource};

use tracing::warn;

use crate::{
    repo::{
        branch::advance_main,
        do_send,
        messages::{AdvanceMain, LoadConfigFromRepo, ValidateRepo},
        RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Handler<AdvanceMain> for RepoActor {
    type Result = ();
    #[tracing::instrument(name = "RepoActor::AdvanceMainTo", skip_all, fields(repo = %self.repo_details, commit = ?msg))]
    fn handle(&mut self, msg: AdvanceMain, ctx: &mut Self::Context) -> Self::Result {
        let Some(repo_config) = self.repo_details.repo_config.clone() else {
            warn!("No config loaded");
            return;
        };
        let Some(open_repository) = &self.open_repository else {
            return;
        };
        let repo_details = self.repo_details.clone();
        let addr = ctx.address();
        let message_token = self.message_token;
        let commit = msg.peel();

        self.update_tui(RepoUpdate::AdvancingMain {
            commit: commit.clone(),
        });

        if let Err(err) = advance_main(commit, &repo_details, &repo_config, &**open_repository) {
            warn!("advance main: {err}");
            self.alert_tui(format!("advance main: {err}"));
        } else {
            self.update_tui(RepoUpdate::MainUpdated);
            if let Some(open_repository) = &self.open_repository {
                match open_repository.fetch() {
                    Ok(()) => self.update_tui_log(git::graph::log(&self.repo_details)),
                    Err(err) => self.alert_tui(format!("fetching: {err}")),
                }
            }
            match repo_config.source() {
                RepoConfigSource::Repo => {
                    do_send(&addr, LoadConfigFromRepo, self.log.as_ref());
                }
                RepoConfigSource::Server => {
                    do_send(&addr, ValidateRepo::new(message_token), self.log.as_ref());
                }
            }
        }
    }
}
