//
use actix::prelude::*;

use git_next_core::git;
use tracing::{warn, Instrument};

use crate::{
    repo::{
        branch::{advance_next, find_next_commit_on_dev},
        do_send,
        messages::{AdvanceNext, AdvanceNextPayload, ValidateRepo},
        RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Handler<AdvanceNext> for RepoActor {
    type Result = ();

    fn handle(&mut self, msg: AdvanceNext, ctx: &mut Self::Context) -> Self::Result {
        let Some(repo_config) = &self.repo_details.repo_config else {
            return;
        };
        let Some(open_repository) = &self.open_repository else {
            return;
        };

        let AdvanceNextPayload {
            next,
            main,
            dev_commit_history,
        } = msg.peel();
        let repo_details = self.repo_details.clone();
        let repo_config = repo_config.clone();
        let addr = ctx.address();

        let (commit, force) = find_next_commit_on_dev(&next, &main, &dev_commit_history);
        if let Some(commit) = &commit {
            self.update_tui(RepoUpdate::AdvancingNext {
                commit: commit.clone(),
                force: force.clone(),
            });
        };
        match advance_next(
            commit,
            force,
            &repo_details,
            repo_config,
            &**open_repository,
            self.message_token,
        ) {
            Ok(message_token) => {
                self.update_tui(RepoUpdate::NextUpdated);
                match open_repository.fetch() {
                    Ok(()) => self.update_tui_log(git::graph::log(&self.repo_details)),
                    Err(err) => self.alert_tui(format!("fetching: {err}")),
                }
                // INFO: pause to allow any CI checks to be started
                let sleep_duration = self.sleep_duration;
                let log = self.log.clone();
                async move {
                    actix_rt::time::sleep(sleep_duration).await;
                    do_send(&addr, ValidateRepo::new(message_token), log.as_ref());
                }
                .in_current_span()
                .into_actor(self)
                .wait(ctx);
            }
            Err(err) => {
                warn!("advance next: {err}");
                self.alert_tui(err.to_string());
            }
        }
    }
}
