//
use actix::prelude::*;

use git_next_core::git::{forge::commit::Status, graph, UserNotification};
use tracing::{debug, Instrument};

use crate::{
    repo::{
        do_send, logger,
        messages::{AdvanceMain, ReceiveCIStatus, ValidateRepo},
        notify_user, RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

impl Handler<ReceiveCIStatus> for RepoActor {
    type Result = ();

    fn handle(&mut self, msg: ReceiveCIStatus, ctx: &mut Self::Context) -> Self::Result {
        logger(self.log.as_ref(), "start: ReceiveCIStatus");
        let (next, status) = msg.peel();
        self.update_tui(RepoUpdate::ReceiveCIStatus {
            status: status.clone(),
        });
        debug!(?status, "");
        let graph_log = graph::log(&self.repo_details);
        self.update_tui_log(graph_log.clone());

        let addr = ctx.address();
        let forge_alias = self.repo_details.forge.forge_alias().clone();
        let repo_alias = self.repo_details.repo_alias.clone();
        let message_token = self.message_token;
        let sleep_duration = self.sleep_duration;
        match status {
            Status::Pass => {
                do_send(&addr, AdvanceMain::new(next), self.log.as_ref());
            }
            Status::Pending => {
                let log = self.log.clone();
                async move {
                    actix_rt::time::sleep(sleep_duration).await;
                    do_send(&addr, ValidateRepo::new(message_token), log.as_ref());
                }
                .in_current_span()
                .into_actor(self)
                .wait(ctx);
            }
            Status::Fail => {
                tracing::warn!("Checks have failed");

                notify_user(
                    self.notify_user_recipient.as_ref(),
                    UserNotification::CICheckFailed {
                        forge_alias,
                        repo_alias,
                        commit: next,
                        log: graph_log,
                    },
                    self.log.as_ref(),
                );
                let log = self.log.clone();
                async move {
                    debug!("sleeping before retrying...");
                    logger(log.as_ref(), "before sleep");
                    actix_rt::time::sleep(sleep_duration).await;
                    logger(log.as_ref(), "after sleep");
                    do_send(&addr, ValidateRepo::new(message_token), log.as_ref());
                }
                .in_current_span()
                .into_actor(self)
                .wait(ctx);
            }
        }
    }
}
