//
use actix::prelude::*;

use tracing::{info, instrument, Instrument as _};

use crate::{
    repo::{
        do_send, logger,
        messages::{AdvanceNext, AdvanceNextPayload, CheckCIStatus, MessageToken, ValidateRepo},
        notify_user, RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

use git_next_core::git::{
    push::Force,
    validation::positions::{validate, Error, Positions},
    UserNotification,
};

impl Handler<ValidateRepo> for RepoActor {
    type Result = ();

    #[instrument(name = "RepoActor::ValidateRepo", skip_all, fields(repo = %self.repo_details, token = %&*msg))]
    fn handle(&mut self, msg: ValidateRepo, ctx: &mut Self::Context) -> Self::Result {
        logger(self.log.as_ref(), "start: ValidateRepo");

        // Message Token - make sure we are only triggered for the latest/current token
        match self.token_status(msg.peel()) {
            TokenStatus::Current => {} // do nothing
            TokenStatus::Expired => {
                logger(
                    self.log.as_ref(),
                    format!("discarded: old message token: {}", self.message_token),
                );
                return; // message is expired
            }
            TokenStatus::New(message_token) => {
                self.message_token = message_token;
                logger(
                    self.log.as_ref(),
                    format!("new message token: {}", self.message_token),
                );
            }
        }
        logger(
            self.log.as_ref(),
            format!("accepted token: {}", self.message_token),
        );

        self.update_tui(RepoUpdate::ValidateRepo);

        // Repository positions
        let Some(ref open_repository) = self.open_repository else {
            logger(self.log.as_ref(), "no open repository");
            self.alert_tui("repo not open");
            return;
        };
        logger(self.log.as_ref(), "have open repository");
        let Some(repo_config) = self.repo_details.repo_config.clone() else {
            logger(self.log.as_ref(), "no repo config");
            self.alert_tui("no repo config");
            return;
        };
        logger(self.log.as_ref(), "have repo config");

        match validate(&**open_repository, &self.repo_details, &repo_config) {
            Ok((
                Positions {
                    main,
                    next,
                    dev,
                    dev_commit_history,
                    next_is_valid,
                },
                git_log,
            )) => {
                info!(%main, %next, %dev, "positions");
                self.update_tui_log(git_log);
                if next_is_valid && next != main {
                    info!("Checking CI");
                    do_send(&ctx.address(), CheckCIStatus::new(next), self.log.as_ref());
                } else if next != dev {
                    info!("Advance next");
                    self.update_tui(RepoUpdate::AdvancingNext {
                        commit: next.clone(),
                        force: Force::No,
                    });
                    do_send(
                        &ctx.address(),
                        AdvanceNext::new(AdvanceNextPayload {
                            next,
                            main,
                            dev_commit_history,
                        }),
                        self.log.as_ref(),
                    );
                } else {
                    info!("do nothing");
                    self.update_tui(RepoUpdate::Okay { main, next, dev });
                }
            }
            Err(Error::Retryable(message)) => {
                info!(?message, "Retryable");
                self.alert_tui(format!("retryable: {message}"));
                logger(self.log.as_ref(), message);
                let addr = ctx.address();
                let message_token = self.message_token;
                let sleep_duration = self.sleep_duration;
                let log = self.log.clone();
                async move {
                    info!("sleeping before retrying...");
                    logger(log.as_ref(), "before sleep");
                    actix_rt::time::sleep(sleep_duration).await;
                    logger(log.as_ref(), "after sleep");
                    do_send(&addr, ValidateRepo::new(message_token), log.as_ref());
                }
                .in_current_span()
                .into_actor(self)
                .wait(ctx);
            }
            Err(Error::UserIntervention(user_notification)) => {
                info!(?user_notification, "User Intervention");
                self.alert_tui(format!("USER INTERVENTION: {user_notification}"));
                if let UserNotification::CICheckFailed { log, .. }
                | UserNotification::DevNotBasedOnMain { log, .. } = &user_notification
                {
                    self.update_tui_log(log.clone());
                }
                notify_user(
                    self.notify_user_recipient.as_ref(),
                    user_notification,
                    self.log.as_ref(),
                );
            }
            Err(Error::NonRetryable(message)) => {
                info!(?message, "NonRetryable");
                self.alert_tui(format!("Error: {message}"));
                logger(self.log.as_ref(), message);
            }
        }
    }
}

enum TokenStatus {
    Current,
    Expired,
    New(MessageToken),
}
impl RepoActor {
    fn token_status(&self, new: MessageToken) -> TokenStatus {
        let current = &self.message_token;
        if &new > current {
            return TokenStatus::New(new);
        }
        if current > &new {
            return TokenStatus::Expired;
        }
        TokenStatus::Current
    }
}
