//
use actix::prelude::*;

use tracing::{info, instrument, warn};

use crate::{
    repo::{
        do_send, logger,
        messages::{ValidateRepo, WebhookNotification},
        ActorLog, RepoActor,
    },
    server::actor::messages::RepoUpdate,
};

use git_next_core::{
    git::{Commit, ForgeLike},
    webhook::{push::Branch, Push},
    BranchName, WebhookAuth,
};

impl Handler<WebhookNotification> for RepoActor {
    type Result = ();

    #[instrument(name = "RepoActor::WebhookMessage", skip_all, fields(token = %self.message_token, repo = %self.repo_details))]
    fn handle(&mut self, msg: WebhookNotification, ctx: &mut Self::Context) -> Self::Result {
        let Some(config) = &self.repo_details.repo_config else {
            logger(self.log.as_ref(), "server has no repo config");
            warn!("No repo config");
            return;
        };
        if validate_notification(
            &msg,
            self.webhook_auth.as_ref(),
            &*self.forge,
            self.log.as_ref(),
        )
        .is_err()
        {
            return;
        }
        let body = msg.body();
        match self.forge.parse_webhook_body(body) {
            Err(err) => {
                logger(self.log.as_ref(), "message parse error - not a push");
                warn!(?err, "Not a 'push'");
                return;
            }
            Ok(push) => match push.branch(config.branches()) {
                None => {
                    logger(self.log.as_ref(), "unknown branch");
                    warn!(
                        ?push,
                        "Unrecognised branch, we should be filtering to only the ones we want"
                    );
                    return;
                }
                Some(Branch::Main) => {
                    self.update_tui(RepoUpdate::WebhookReceived {
                        branch: Branch::Main,
                        push: push.clone(),
                    });
                    if handle_push(
                        push,
                        &config.branches().main(),
                        &mut self.last_main_commit,
                        self.log.as_ref(),
                    )
                    .is_err()
                    {
                        return;
                    };
                }
                Some(Branch::Next) => {
                    self.update_tui(RepoUpdate::WebhookReceived {
                        branch: Branch::Next,
                        push: push.clone(),
                    });
                    if handle_push(
                        push,
                        &config.branches().next(),
                        &mut self.last_next_commit,
                        self.log.as_ref(),
                    )
                    .is_err()
                    {
                        return;
                    };
                }
                Some(Branch::Dev) => {
                    self.update_tui(RepoUpdate::WebhookReceived {
                        branch: Branch::Dev,
                        push: push.clone(),
                    });
                    if handle_push(
                        push,
                        &config.branches().dev(),
                        &mut self.last_dev_commit,
                        self.log.as_ref(),
                    )
                    .is_err()
                    {
                        return;
                    };
                }
            },
        }
        let message_token = self.message_token.next();
        info!(
            token = %message_token,
            "New commit"
        );
        do_send(
            &ctx.address(),
            ValidateRepo::new(message_token),
            self.log.as_ref(),
        );
    }
}

fn validate_notification(
    msg: &WebhookNotification,
    webhook_auth: Option<&WebhookAuth>,
    forge: &dyn ForgeLike,
    log: Option<&ActorLog>,
) -> Result<(), ()> {
    let Some(expected_authorization) = webhook_auth else {
        logger(log, "server has no auth token");
        warn!("Don't know what authorization to expect");
        return Err(());
    };

    if !forge.is_message_authorised(msg, expected_authorization) {
        logger(log, "message authorisation is invalid");
        warn!(
            "Invalid authorization - expected {}",
            expected_authorization
        );
        return Err(());
    }
    if forge.should_ignore_message(msg) {
        logger(log, "forge sent ignorable message");
        return Err(());
    }
    Ok(())
}

fn handle_push(
    push: Push,
    branch: &BranchName,
    last_commit: &mut Option<Commit>,
    log: Option<&ActorLog>,
) -> Result<(), ()> {
    logger(log, format!("message is for {branch} branch"));
    let commit = Commit::from(push);
    if last_commit.as_ref() == Some(&commit) {
        logger(log, format!("not a new commit on {branch}"));
        info!(
            %branch ,
            %commit,
            "Ignoring - already aware of branch at commit",
        );
        return Err(());
    }
    last_commit.replace(commit);
    Ok(())
}
