//
use actix::prelude::*;

use git_next_core::{ForgeAlias, RepoAlias};
use tracing::info;

use crate::{
    alerts::messages::UpdateShout,
    repo::{messages::CloneRepo, RepoActor},
    server::actor::{
        messages::{ReceiveValidAppConfig, ServerUpdate, ValidAppConfig},
        ServerActor,
    },
    webhook::{
        messages::ShutdownWebhook,
        router::{AddWebhookRecipient, WebhookRouterActor},
        WebhookActor,
    },
};

impl Handler<ReceiveValidAppConfig> for ServerActor {
    type Result = ();

    fn handle(&mut self, msg: ReceiveValidAppConfig, ctx: &mut Self::Context) -> Self::Result {
        let ValidAppConfig {
            app_config,
            socket_address,
            storage: server_storage,
        } = msg.peel();
        // shutdown any existing webhook actor
        if let Some(webhook_actor_addr) = self.webhook_actor_addr.take() {
            webhook_actor_addr.do_send(ShutdownWebhook);
        }
        self.generation.inc();
        // Webhook Server
        info!("Starting Webhook Server...");
        let webhook_router = WebhookRouterActor::default().start();
        let listen_url = app_config.listen().url();
        let notify_user_recipient = self.alerts.clone().recipient();
        let server_addr = Some(ctx.address());
        // Forge Actors
        for (forge_alias, forge_config) in app_config.forges() {
            let repo_actors = self
                .create_forge_repos(
                    forge_config,
                    forge_alias.clone(),
                    &server_storage,
                    listen_url,
                    &notify_user_recipient,
                    server_addr.clone(),
                )
                .into_iter()
                .map(start_repo_actor)
                .collect::<Vec<_>>();
            repo_actors
                .iter()
                .map(|(repo_alias, addr)| {
                    AddWebhookRecipient::new(
                        forge_alias.clone(),
                        repo_alias.clone(),
                        addr.clone().recipient(),
                    )
                })
                .for_each(|msg| webhook_router.do_send(msg));
            for (repo_alias, addr) in repo_actors {
                self.repo_actors
                    .insert((forge_alias.clone(), repo_alias), addr);
            }
        }
        let webhook_actor_addr =
            WebhookActor::new(socket_address, webhook_router.recipient()).start();
        self.webhook_actor_addr.replace(webhook_actor_addr);
        let shout = app_config.shout().clone();
        self.app_config.replace(app_config.clone());
        self.do_send(
            ServerUpdate::AppConfigLoaded {
                app_config: ValidAppConfig {
                    app_config,
                    socket_address,
                    storage: server_storage,
                },
            },
            ctx,
        );
        self.alerts.do_send(UpdateShout::new(shout));
    }
}

fn start_repo_actor(actor: (ForgeAlias, RepoAlias, RepoActor)) -> (RepoAlias, Addr<RepoActor>) {
    let (forge_name, repo_alias, actor) = actor;
    let span = tracing::info_span!("start_repo_actor", forge = %forge_name, repo = %repo_alias);
    let _guard = span.enter();
    let addr = actor.start();
    addr.do_send(CloneRepo);
    tracing::info!("Started");
    (repo_alias, addr)
}
