//
use actix::prelude::*;
use messages::{ReceiveAppConfig, ServerUpdate, Shutdown};
use tracing::error;

#[cfg(test)]
mod tests;

mod handlers;
pub mod messages;

use crate::{
    alerts::messages::NotifyUser, alerts::AlertsActor, forge::Forge, repo::RepoActor,
    webhook::WebhookActor,
};

use git_next_core::{
    git::{repository::factory::RepositoryFactory, Generation, RepoDetails},
    server::{self, AppConfig, ListenUrl, Storage},
    ForgeAlias, ForgeConfig, GitDir, RepoAlias, ServerRepoConfig, StoragePathType,
};

use kxio::{fs::FileSystem, network::Network};

use std::{
    collections::BTreeMap,
    path::PathBuf,
    sync::{Arc, RwLock},
};

#[derive(Debug, derive_more::Display, derive_more::From)]
pub enum Error {
    #[display("Failed to create data directories")]
    FailedToCreateDataDirectory(kxio::fs::Error),

    #[display("The forge data path is not a directory: {path:?}")]
    ForgeDirIsNotDirectory {
        path: PathBuf,
    },

    Config(server::Error),

    Io(std::io::Error),
}
type Result<T> = core::result::Result<T, Error>;

#[allow(clippy::module_name_repetitions)]
#[derive(derive_with::With)]
#[with(message_log)]
pub struct ServerActor {
    app_config: Option<AppConfig>,
    generation: Generation,
    webhook_actor_addr: Option<Addr<WebhookActor>>,
    fs: FileSystem,
    net: Network,
    alerts: Addr<AlertsActor>,
    repository_factory: Box<dyn RepositoryFactory>,
    sleep_duration: std::time::Duration,
    repo_actors: BTreeMap<(ForgeAlias, RepoAlias), Addr<RepoActor>>,

    shutdown_trigger: Option<std::sync::mpsc::Sender<String>>,
    subscribers: Vec<Recipient<ServerUpdate>>,

    // testing
    message_log: Option<Arc<RwLock<Vec<String>>>>,
}
impl Actor for ServerActor {
    type Context = Context<Self>;
}

impl ServerActor {
    pub fn new(
        fs: FileSystem,
        net: Network,
        alerts: Addr<AlertsActor>,
        repo: Box<dyn RepositoryFactory>,
        sleep_duration: std::time::Duration,
    ) -> Self {
        let generation = Generation::default();
        Self {
            app_config: None,
            generation,
            webhook_actor_addr: None,
            fs,
            net,
            alerts,
            repository_factory: repo,
            shutdown_trigger: None,
            subscribers: Vec::default(),
            sleep_duration,
            repo_actors: BTreeMap::new(),
            message_log: None,
        }
    }
    fn create_forge_data_directories(
        &self,
        app_config: &AppConfig,
        server_dir: &std::path::Path,
    ) -> Result<()> {
        for (forge_name, _forge_config) in app_config.forges() {
            let forge_dir: PathBuf = (&forge_name).into();
            let path = server_dir.join(&forge_dir);
            if self.fs.path_exists(&path)? {
                if !self.fs.path_is_dir(&path)? {
                    return Err(Error::ForgeDirIsNotDirectory { path });
                }
            } else {
                tracing::info!(%forge_name, ?path, "creating storage");
                self.fs.dir_create_all(&path)?;
            }
        }

        Ok(())
    }

    fn create_forge_repos(
        &self,
        forge_config: &ForgeConfig,
        forge_name: ForgeAlias,
        server_storage: &Storage,
        listen_url: &ListenUrl,
        notify_user_recipient: &Recipient<NotifyUser>,
        server_addr: Option<Addr<Self>>,
    ) -> Vec<(ForgeAlias, RepoAlias, RepoActor)> {
        let span =
            tracing::info_span!("create_forge_repos", name = %forge_name, config = %forge_config);

        let _guard = span.enter();
        tracing::info!("Creating Forge");
        let mut repos = vec![];
        let creator = self.create_actor(
            forge_name,
            forge_config.clone(),
            server_storage,
            listen_url,
            server_addr,
        );
        for (repo_alias, server_repo_config) in forge_config.repos() {
            let forge_repo = creator((
                repo_alias,
                server_repo_config,
                notify_user_recipient.clone(),
            ));
            tracing::info!(
                alias = %forge_repo.1,
                "Created Repo"
            );
            repos.push(forge_repo);
        }
        repos
    }

    fn create_actor(
        &self,
        forge_name: ForgeAlias,
        forge_config: ForgeConfig,
        server_storage: &Storage,
        listen_url: &ListenUrl,
        server_addr: Option<Addr<Self>>,
    ) -> impl Fn(
        (RepoAlias, &ServerRepoConfig, Recipient<NotifyUser>),
    ) -> (ForgeAlias, RepoAlias, RepoActor) {
        let server_storage = server_storage.clone();
        let listen_url = listen_url.clone();
        let net = self.net.clone();
        let repository_factory = self.repository_factory.duplicate();
        let generation = self.generation;
        let sleep_duration = self.sleep_duration;
        move |(repo_alias, server_repo_config, notify_user_recipient)| {
            let span = tracing::info_span!("create_actor", alias = %repo_alias, config = %server_repo_config);
            let _guard = span.enter();
            tracing::info!("Creating Repo");
            let gitdir = server_repo_config.gitdir().map_or_else(
                || {
                    GitDir::new(
                        server_storage
                            .path()
                            .join(forge_name.to_string())
                            .join(repo_alias.to_string()),
                        StoragePathType::Internal,
                    )
                },
                |gitdir| gitdir,
            );
            // INFO: can't canonicalise gitdir as the path needs to exist to do that and we may not
            //       have cloned the repo yet
            let repo_details = RepoDetails::new(
                generation,
                &repo_alias,
                server_repo_config,
                &forge_name,
                &forge_config,
                gitdir,
            );
            let forge = Forge::create(repo_details.clone(), net.clone());
            tracing::info!("Starting Repo Actor");
            let actor = RepoActor::new(
                repo_details,
                forge,
                listen_url.clone(),
                generation,
                net.clone(),
                repository_factory.duplicate(),
                sleep_duration,
                Some(notify_user_recipient),
                server_addr.clone(),
            );
            (forge_name.clone(), repo_alias, actor)
        }
    }

    fn server_storage(&self, app_config: &ReceiveAppConfig) -> Option<Storage> {
        let server_storage = app_config.storage().clone();
        let dir = server_storage.path();
        if !dir.exists() {
            if let Err(err) = self.fs.dir_create(dir) {
                error!(?err, ?dir, "Failed to create server storage");
                return None;
            }
        }
        let Ok(canon) = dir.canonicalize() else {
            error!(?dir, "Failed to confirm server storage");
            return None;
        };
        if let Err(err) = self.create_forge_data_directories(app_config, &canon) {
            error!(?err, "Failure creating forge storage");
            return None;
        }
        Some(server_storage)
    }

    /// Attempts to gracefully shutdown the server before stopping the system.
    fn abort(&mut self, ctx: &<Self as actix::Actor>::Context, message: impl Into<String>) {
        self.do_send(crate::server::actor::messages::Shutdown, ctx);
        if let Some(t) = self.shutdown_trigger.take() {
            let _ = t.send(message.into());
        } else {
            error!("{}", message.into());
            self.do_send(Shutdown, ctx);
            // System::current().stop_with_code(1);
        }
    }

    fn do_send<M>(&self, msg: M, ctx: &<Self as actix::Actor>::Context)
    where
        M: actix::Message + Send + 'static + std::fmt::Debug,
        Self: actix::Handler<M>,
        <M as actix::Message>::Result: Send,
    {
        if let Some(message_log) = &self.message_log {
            let log_message = format!("send: {msg:?}");
            if let Ok(mut log) = message_log.write() {
                log.push(log_message);
            }
        }
        if cfg!(not(test)) {
            ctx.address().do_send(msg);
        }
    }
}
