//
pub mod actor;

#[cfg(test)]
mod tests;

use actix::prelude::*;
use actix_rt::signal;
use actor::messages::ShutdownTrigger;

use crate::{
    alerts::{AlertsActor, History},
    file_watcher::{watch_file, FileUpdated},
};

#[allow(clippy::module_name_repetitions)]
pub use actor::ServerActor;

use git_next_core::git::RepositoryFactory;

use color_eyre::{eyre::Context, Result};
use kxio::{fs::FileSystem, network::Network};
use tracing::info;

use std::{
    path::PathBuf,
    sync::{atomic::Ordering, mpsc::channel, Arc, RwLock},
    time::Duration,
};

const A_DAY: Duration = Duration::from_secs(24 * 60 * 60);

pub fn init(fs: &FileSystem) -> Result<()> {
    let file_name = "git-next-server.toml";
    let pathbuf = PathBuf::from(file_name);
    if fs
        .path_exists(&pathbuf)
        .with_context(|| format!("Checking for existing file: {pathbuf:?}"))?
    {
        eprintln!("The configuration file already exists at {pathbuf:?} - not overwritting it.",);
    } else {
        fs.file_write(&pathbuf, include_str!("server-default.toml"))
            .with_context(|| format!("Writing file: {pathbuf:?}"))?;
        println!("Created a default configuration file at {pathbuf:?}",);
    }
    Ok(())
}

#[allow(clippy::too_many_lines)]
pub fn start(
    ui: bool,
    fs: FileSystem,
    net: Network,
    repo: Box<dyn RepositoryFactory>,
    sleep_duration: std::time::Duration,
) -> Result<()> {
    if ui {
        #[cfg(feature = "tui")]
        {
            crate::tui::logging::initialize_logging()?;
        }
    } else {
        init_logging();
    }

    let shutdown_message_holder: Arc<RwLock<Option<String>>> = Arc::new(RwLock::new(None));
    let shutdown_message_holder_exec = shutdown_message_holder.clone();
    let file_watcher_err_holder: Arc<RwLock<Option<anyhow::Error>>> = Arc::new(RwLock::new(None));
    let file_watcher_err_holder_exec = file_watcher_err_holder.clone();
    let execution = async move {
        info!("Starting Alert Dispatcher...");
        let alerts_addr = AlertsActor::new(None, History::new(A_DAY), net.clone()).start();

        info!("Starting Server...");
        let server =
            ServerActor::new(fs.clone(), net.clone(), alerts_addr, repo, sleep_duration).start();

        info!("Starting File Watcher...");
        let watch_file = watch_file("git-next-server.toml".into(), server.clone().recipient());
        let fw_shutdown = match watch_file {
            Ok(fw_shutdown) => fw_shutdown,
            Err(err) => {
                // shutdown now
                server.do_send(crate::server::actor::messages::Shutdown);
                actix_rt::time::sleep(std::time::Duration::from_millis(10)).await;
                System::current().stop();
                let _ = file_watcher_err_holder_exec
                    .write()
                    .map(|mut o| o.replace(err));
                return;
            }
        };

        let (tx_shutdown, rx_shutdown) = channel::<String>();
        if ui {
            #[cfg(feature = "tui")]
            {
                use crate::server::actor::messages::SubscribeToUpdates;
                use crate::tui;

                let tui_addr = tui::Tui::new(tx_shutdown.clone()).start();
                server.do_send(SubscribeToUpdates::new(tui_addr.clone().recipient()));
                server.do_send(ShutdownTrigger::new(tx_shutdown));
                server.do_send(FileUpdated); // update file after ui subscription in place
                loop {
                    let _ = tui_addr.send(tui::Tick).await;
                    if let Ok(message) = rx_shutdown.try_recv() {
                        let _ = shutdown_message_holder_exec
                            .write()
                            .map(|mut o| o.replace(message));
                        break;
                    }
                    actix_rt::time::sleep(Duration::from_millis(16)).await;
                }
            }
        } else {
            server.do_send(ShutdownTrigger::new(tx_shutdown.clone()));
            server.do_send(FileUpdated);

            info!("Server running - Press Ctrl-C to stop...");
            tokio::select! {
                _r = signal::ctrl_c() => {
                    info!("Ctrl-C received, shutting down...");
                }
                _x = async move {
                    loop{
                        if let Ok(message) = rx_shutdown.try_recv() {
                            let _ = shutdown_message_holder_exec
                                .write()
                                .map(|mut o| o.replace(message));
                            break;
                        }
                        actix_rt::task::yield_now().await;
                    }
                } => {
                    info!("signaled shutdown");
                }
            };
        }

        // shutdown
        fw_shutdown.store(true, Ordering::Relaxed);
        server.do_send(crate::server::actor::messages::Shutdown);
        actix_rt::time::sleep(std::time::Duration::from_millis(10)).await;
        System::current().stop();
    };

    let system = System::new();
    Arbiter::current().spawn(execution);
    system.run()?;

    // check for error from server thread
    #[allow(clippy::unwrap_used)]
    if let Some(err) = &*shutdown_message_holder.read().unwrap() {
        #[cfg(feature = "tui")]
        if ui {
            ratatui::restore();
        }
        if !err.is_empty() {
            return Err(color_eyre::eyre::eyre!(format!("{err}")));
        }
    }

    // check for error from file watcher thread
    #[allow(clippy::unwrap_used)]
    if let Some(err) = &*file_watcher_err_holder.read().unwrap() {
        #[cfg(feature = "tui")]
        if ui {
            ratatui::restore();
        }
        return Err(color_eyre::eyre::eyre!(format!("{err}")));
    }

    Ok(())
}

pub fn init_logging() {
    use tracing::Level;
    use tracing_subscriber::FmtSubscriber;

    let subscriber = FmtSubscriber::builder()
        .with_target(false)
        .with_file(true)
        .with_line_number(true)
        .with_max_level(Level::INFO)
        .finish();

    #[allow(clippy::expect_used)]
    tracing::subscriber::set_global_default(subscriber).expect("setting default subscriber failed");
}
