//
mod handlers;
pub mod messages;
mod model;

#[cfg(test)]
mod tests;

use std::sync::mpsc::Sender;

use actix::{Actor, ActorContext as _, Context};

pub use model::*;

use ratatui::{
    crossterm::event::{self, KeyCode, KeyEventKind},
    DefaultTerminal,
};
use tui_scrollview::ScrollViewState;

#[derive(Debug)]
pub struct Tui {
    terminal: Option<DefaultTerminal>,
    signal_shutdown: Sender<String>,
    pub state: State,
    scroll_view_state: ScrollViewState,
}
impl Actor for Tui {
    type Context = Context<Self>;
    fn started(&mut self, _ctx: &mut Self::Context) {
        self.terminal.replace(ratatui::init());
    }
    fn stopped(&mut self, _ctx: &mut Self::Context) {
        self.terminal.take();
        ratatui::restore();
    }
}
impl Tui {
    pub fn new(signal_shutdown: Sender<String>) -> Self {
        Self {
            terminal: None,
            signal_shutdown,
            state: State::initial(),
            scroll_view_state: ScrollViewState::default(),
        }
    }

    fn draw(&mut self) -> std::io::Result<()> {
        let t = self.terminal.take();
        let scroll_view_state = &mut self.scroll_view_state;
        let state = &self.state;
        if let Some(mut terminal) = t {
            terminal.draw(|frame| {
                frame.render_stateful_widget(state, frame.area(), scroll_view_state);
            })?;
            self.terminal = Some(terminal);
        } else {
            eprintln!("No terminal setup");
        }
        Ok(())
    }

    fn handle_input(&mut self, ctx: &mut <Self as actix::Actor>::Context) -> std::io::Result<()> {
        if event::poll(std::time::Duration::from_millis(16))? {
            let event::Event::Key(key) = event::read()? else {
                return Ok(());
            };
            if key.kind != KeyEventKind::Press {
                return Ok(());
            }
            match key.code {
                KeyCode::Char('q') => {
                    ctx.stop();
                    if let Err(err) = self.signal_shutdown.send(String::new()) {
                        tracing::error!(?err, "Failed to signal shutdown");
                    }
                }
                KeyCode::Char('j') | KeyCode::Down => self.scroll_view_state.scroll_down(),
                KeyCode::Char('k') | KeyCode::Up => self.scroll_view_state.scroll_up(),
                KeyCode::Char('f') | KeyCode::PageDown => {
                    self.scroll_view_state.scroll_page_down();
                }
                KeyCode::Char('b') | KeyCode::PageUp => {
                    self.scroll_view_state.scroll_page_up();
                }
                KeyCode::Char('g') | KeyCode::Home => {
                    self.scroll_view_state.scroll_to_top();
                }
                KeyCode::Char('G') | KeyCode::End => {
                    self.scroll_view_state.scroll_to_bottom();
                }
                _ => (),
            }
        }
        Ok(())
    }
}
